#!/usr/bin/env python

from doit.task import clean_targets
from doit.tools import LongRunning
import glob
from itertools import count
import numpy as np
import os
import pandas as pd

from .profile import profile_task
from .util import create_doit_task as doit_task
from .util import which, parallel_fasta, title

LASTAL_CFG = { "params": [],
               "frameshift": 15 }
LASTDB_CFG = { "params": ["-w3"] }

def clean_lastdb(db_prefix):
    files = glob.glob('{0}.*'.format(db_prefix))
    for fn in files:
        try:
            os.remove(fn)
        except OSError as e:
            pass


@doit_task
@profile_task
def lastdb_task(db_fn, db_out_prefix=None, prot=True,
                params=LASTDB_CFG['params'], task_dep=None):
    '''Create a pydoit task to run lastdb.

    WARNING: This does not define a file_dep, to make sure it doesn't
    get executed when the dependency and targets already exist. This means
    that if the db_fn is created by another task, it MUST be defined before the
    lastdb task. This is a kludge which will eventually be fixed...

    Args:
        db_fn (str): The FASTA file to format.
        db_out_prefix (str): Prefix for the database files. Same as db_fn
                             if None (default).
        prot (bool): True if a protein FASTA, False otherwise.
        params (list): A list of additional parameters.
    Returns:
        dict: A pydoit task.
    '''

    cmd = [which('lastdb')]
    if prot:
        cmd.append('-p')
    if params is not None:
        cmd.extend([str(p) for p in params])
    if db_out_prefix is None:
        db_out_prefix = db_fn
    cmd.extend([db_out_prefix, db_fn])
    cmd = ' '.join(cmd)

    name = 'lastdb:' + os.path.basename(db_out_prefix)

    task_d =  {'name': name,
              'title': title,
              'actions': [cmd],
              'targets': ['{0}.prj'.format(db_out_prefix)],
              'uptodate': [True],
              'clean': [clean_targets,
                        (clean_lastdb, [db_out_prefix])]}
    if task_dep is not None:
        task_d['task_dep'] = task_dep

    return task_d


@doit_task
@profile_task
def lastal_task(query, db, out_fn, translate=False,
                frameshift=LASTAL_CFG['frameshift'], cutoff=0.00001, 
                n_threads=1, pbs=None, params=None):
    '''Create a pydoit task to run lastal

    Args:
        query (str): The file with the query sequences.
        db (str): The database file prefix.
        out_fn (str): Destination file for alignments.
        translate (bool): True if query is a nucleotide FASTA.
        frameshift (int): Frameshift penalty for translated alignment.
        n_threads (int): Number of threads to run with.
    Returns:
        dict: A pydoit task.
    '''

    lastal_exc = which('lastal')
    name = 'lastal:{0}'.format(os.path.join(out_fn))

    cmd = [lastal_exc]
    if translate:
        cmd.append('-F' + str(frameshift))
    if cutoff is not None:
        cutoff = round(1.0 / cutoff, 2)
        cmd.append('-D' + str(cutoff))
    if params is not None:
        cmd.extend(params)
    cmd.append(db)

    cmd =  parallel_fasta(query, out_fn, cmd, n_threads, pbs=pbs)

    return {'name': name,
            'title': title,
            'actions': [cmd],
            'targets': [out_fn],
            'file_dep': [query, db + '.prj'],
            'clean': [clean_targets]}


def next_or_raise(fp):
    counter = count()
    def func(raise_exc=True):
        line = fp.readline()
        n = next(counter)
        if raise_exc is True and line == '':
            raise RuntimeError('Malformed MAF file (line {0})'.format(n))
        return line
    return func


class MafParser(object):

    def __init__(self, filename, aln_strings=False, chunksize=10000, **kwargs):
        '''Parser for LAST's MAF output.

        Args:
            filename (str): The MAF file.
            aln_strings (bool): If True, parse the alignment strings as well.
            chunksize (int): Size of chunks to parse.
        '''
        self.chunksize = chunksize
        self.filename = filename
        self.aln_strings = aln_strings
        self.LAMBDA = None
        self.K = None

    def read(self):
        '''Read the entire file at once and return as a single DataFrame.
        '''
        return pd.concat(iter(self), ignore_index=True)

    def __iter__(self):
        '''Iterator yielding DataFrames of length chunksize holding MAF alignments.

        An extra column is added for bitscore, using the equation described here:
            http://last.cbrc.jp/doc/last-evalues.html

        Args:
            fn (str): Path to the MAF alignment file.
            chunksize (int): Alignments to parse per iteration.
        Yields:
            DataFrame: Pandas DataFrame with the alignments.
        '''
        data = []
        with open(self.filename) as fp:
            guarded_next = next_or_raise(fp)
            n = 0
            while (True):
                line = guarded_next(raise_exc=False)
                if line == '':
                    break
                line = line.strip()
                if not line:
                    continue
                if line.startswith('#'):
                    if 'lambda' in line:
                        meta = line.strip(' #').split()
                        meta = {k:v for k, _, v in map(lambda x: x.partition('='), meta)}
                        self.LAMBDA = float(meta['lambda'])
                        self.K = float(meta['K'])
                    else:
                        continue
                if line.startswith('a'):
                    cur_aln = {}

                    # Alignment info
                    tokens = line.split()
                    for token in tokens[1:]:
                        key, _, val = token.strip().partition('=')
                        cur_aln[key] = float(val)

                    # First sequence info
                    line = guarded_next()
                    tokens = line.split()
                    cur_aln['s_name'] = tokens[1]
                    cur_aln['s_start'] = int(tokens[2])
                    cur_aln['s_aln_len'] = int(tokens[3])
                    cur_aln['s_strand'] = tokens[4]
                    cur_aln['s_len'] = int(tokens[5])
                    if self.aln_strings:
                        cur_aln['s_aln'] = tokens[6]

                    # First sequence info
                    line = guarded_next() 
                    tokens = line.split()
                    cur_aln['q_name'] = tokens[1]
                    cur_aln['q_start'] = int(tokens[2])
                    cur_aln['q_aln_len'] = int(tokens[3])
                    cur_aln['q_strand'] = tokens[4]
                    cur_aln['q_len'] = int(tokens[5])
                    if self.aln_strings:
                        cur_aln['q_aln'] = tokens[6]

                    data.append(cur_aln)
                    if len(data) >= self.chunksize:
                        if self.LAMBDA is None:
                            raise RuntimeError("old version of lastal; please update")
                        yield self._build_df(data)
                        data = []

        if data:
            yield self._build_df(data)

    def _build_df(self, data):

        def _fix_sname(name):
            new, _, _ = name.partition(',')
            return new

        df = pd.DataFrame(data)
        df['s_name'] = df['s_name'].apply(_fix_sname)
        setattr(df, 'LAMBDA', self.LAMBDA)
        setattr(df, 'K', self.K)
        df['bitscore'] = (self.LAMBDA * df['score'] - np.log(self.K)) / np.log(2)

        return df





import pytest
from os import path

from shmlast.tests.utils import runscript, datadir

def test_rbl(tmpdir, datadir):
    query = datadir('sacPom.cdna.fa')
    database = datadir('sacPom.pep.fa')
    
    print(query, database, tmpdir)
    args = ['rbl', '-q', query, '-d', database]
    runscript('shmlast', args, directory=str(tmpdir))

    assert tmpdir.ensure('sacPom.cdna.fa.x.sacPom.pep.fa.rbl.csv')

def test_crbl(tmpdir, datadir):
    query = datadir('sacPom.cdna.fa')
    database = datadir('sacPom.pep.fa')
    out_fn = str(tmpdir.join('test.csv'))
    
    args = ['crbl', '-q', query, '-d', database, '-o', out_fn]
    status, out, err = runscript('shmlast', args, directory=str(tmpdir))
    assert status == 0
    assert tmpdir.ensure(out_fn)



import pytest
import pandas as pd

from shmlast.tests.utils import datadir
from shmlast.hits import BestHits


def check_df_equals(dfA, dfB, col='E'):
    dfA = dfA.sort_values(col, inplace=False).sort_index(axis=1)
    dfA.reset_index(inplace=True, drop=True)

    dfB = dfB.sort_values(col, inplace=False).sort_index(axis=1)
    dfB.reset_index(inplace=True, drop=True)

    print('Checking DataFrame Equality:\n', dfA)
    print(dfB)

    return dfA.equals(dfB)


def test_besthits_inplace(datadir):
    '''Test BestHits.best_hits with inplace=True
    '''
    input_df = pd.read_csv(datadir('query.maf.csv'))
    expected_df = pd.read_csv(datadir('besthits.expected.csv'))

    bh = BestHits()
    bh.best_hits(input_df, inplace=True)
    
    assert check_df_equals(expected_df, input_df)


def test_besthits_non_inplace(datadir):
    '''Test BestHits.best_hits with inplace=False
    '''
    input_df = pd.read_csv(datadir('query.maf.csv'))
    expected_df = pd.read_csv(datadir('besthits.expected.csv'))

    bh = BestHits()
    results_df = bh.best_hits(input_df, inplace=False)
    
    assert check_df_equals(expected_df, results_df)


def test_reciprocal_best_hits(datadir):
    query_df = pd.read_csv(datadir('query.maf.csv'))
    db_df = pd.read_csv(datadir('db.maf.csv'))
    expected_df = pd.read_csv(datadir('reciprocals.expected.csv'))

    bh = BestHits()
    results_df = bh.reciprocal_best_hits(query_df, db_df, inplace=False)
    
    assert check_df_equals(results_df, expected_df)



#!/usr/bin/env python
from __future__ import print_function

import pytest

import json
import os
import sys

from shmlast.tests.utils import datadir, run_task, run_tasks, check_status, touch
from shmlast.last import lastal_task
from shmlast.last import lastdb_task
from shmlast.last import MafParser

LASTDB_EXTENSIONS = ['.bck', '.des', '.prj', '.sds', '.ssp', '.suf', '.tis']



def test_lastdb_task_nucl(tmpdir, datadir):
    with tmpdir.as_cwd():
        tf = datadir('test-transcript.fa')

        task = lastdb_task(tf, tf, prot=False)
        run_tasks([task], ['run'])
        status = check_status(task)
        print('PATH:', os.environ['PATH'], file=sys.stderr)

        for ext in LASTDB_EXTENSIONS:
            assert os.path.isfile(tf + ext)

        assert status.status == 'up-to-date'


def test_lastdb_task_prot(tmpdir, datadir):
    with tmpdir.as_cwd():
        tf = datadir('test-protein.fa')

        task = lastdb_task(tf, tf, prot=True)
        run_tasks([task], ['run'])
        status = check_status(task)
        
        for ext in LASTDB_EXTENSIONS:
            assert os.path.isfile(tf + ext)

        assert status.status == 'up-to-date'


def test_lastdb_task_existing(tmpdir, datadir):
    with tmpdir.as_cwd():
        tf = datadir('test-protein.fa')
        for ext in LASTDB_EXTENSIONS:
            touch(tf + ext)

        task = lastdb_task(tf, tf, prot=True)
        run_tasks([task], ['run'])
        print(task, file=sys.stderr)
        status = check_status(task)

        assert status.status == 'up-to-date'


def test_lastal_task_nucl_x_prot(tmpdir, datadir):
    with tmpdir.as_cwd():
        prot = datadir('test-protein.fa')
        tr = datadir('test-transcript.fa')
        out = tmpdir.join('test-out').strpath

        db_task = lastdb_task(prot, prot)
        aln_task = lastal_task(tr, prot, out,  
                                translate=True, 
                                cutoff=None)
        run_tasks([db_task, aln_task], ['run'])

        aln = ''.join(open(out).readlines())
        print(aln, file=sys.stderr)

        assert 'SPAC212_RecQ_type_DNA_helicase_PROTEIN' in aln
        assert 'SPAC212_RecQ_type_DNA_helicase_TRANSCRIPT' in aln
        assert 'lambda' in aln, 'lambda missing, wrong LAST version?'
        

def test_lastal_task_prot_x_prot(tmpdir, datadir):
    with tmpdir.as_cwd():
        prot = datadir('test-protein.fa')
        out = tmpdir.join('test-out').strpath
            
        db_task = lastdb_task(prot, prot)
        aln_task = lastal_task(prot, prot, out,
                                translate=False,
                                cutoff=None)
        run_tasks([db_task, aln_task], ['run'])

        aln = ''.join(open(out).readlines())
        print(aln, file=sys.stderr)

        assert aln.count('SPAC212_RecQ_type_DNA_helicase_PROTEIN') == 2
        assert 'EG2=0' in aln
        assert 'E=0' in aln
        assert 'lambda' in aln, 'lambda missing, wrong LAST version?'

def test_lastal_task_multithreaded(tmpdir, datadir):
    with tmpdir.as_cwd():
        for n_threads in (3,4,5):
            prot = datadir('test-protein.fa')
            tr = datadir('pom.50.fa')
            out_single = tmpdir.join('out-single').strpath
            out_multi = tmpdir.join('out-multi').strpath

            db_task = lastdb_task(prot, prot)
            aln_task_single = lastal_task(tr, prot, out_single, 
                                           translate=True, 
                                           cutoff=None)

            aln_task_multi = lastal_task(tr, prot, out_multi,
                                         translate=True, 
                                         cutoff=None,
                                         n_threads=n_threads)
            run_tasks([db_task, aln_task_multi, aln_task_single], 
                      ['run'])

            alns_single = MafParser(out_single).read()
            alns_multi = MafParser(out_multi).read()

            assert all(alns_single['E'].sort_values() == \
                       alns_multi['E'].sort_values())

def test_lastal_task_uptodate(tmpdir, datadir):
    with tmpdir.as_cwd():
        prot = datadir('test-protein.fa')
        out = tmpdir.join('test-out').strpath

        db_task = lastdb_task(prot, prot)
        aln_task = lastal_task(prot, prot, out,
                                translate=False,
                                cutoff=None)
        # Run it once
        run_tasks([db_task, aln_task], ['run'])
        # Now run again and check the status
        #run_tasks(aln_tasks, ['run'])
        print(aln_task)
        status = check_status(aln_task, tasks=[aln_task, db_task])
        assert status.status == 'up-to-date'



#!/usr/bin/env python

import hashlib
import os
from string import digits

from doit.tools import run_once, create_folder, title_with_actions, LongRunning
from doit.tools import PythonInteractiveAction
from doit.task import clean_targets, dict_to_task
from doit.cmd_base import TaskLoader
from doit.doit_cmd import DoitMain


def leftpad(s):
    return '\n'.join('    {0}'.format(i) for i in s.split('\n'))


def title(task):
    """return task name task actions"""
    if task.actions:
        title = "\n\t".join([leftpad(str(action)) for action in task.actions])
    # A task that contains no actions at all
    # is used as group task
    else:
        title = "Group: %s" % ", ".join(task.task_dep)
    return "%s => \n%s"% (task.name, title)


def create_doit_task(task_dict_func):
    '''Wrapper to decorate functions returning pydoit
    Task dictionaries and have them return pydoit Task
    objects
    '''
    def d_to_t(*args, **kwargs):
        ret_dict = task_dict_func(*args, **kwargs)
        return dict_to_task(ret_dict)
    return d_to_t


def task_str(task):
    return '{{ Task: {0}\n  actions: {1}\n  file_dep: {2}'\
           '\n  task_dep: {3}\n  targets: {4} }}'.format(task.name, task.actions,
                                                         task.file_dep, task.task_dep,
                                                         task.targets)


def prog_string(subcommand, version, action):
    s = ['shmlast {0} -- Camille Scott, 2016'.format(version)]
    s.append('\n')
    s.append(len(s[0]) * '-')
    s.append('\n')
    s.append('subcommand: {0}'.format(subcommand))
    s.append('\n')
    s.append('doit action: {0}'.format(action))
    return ''.join(s)


class ShortenedPythonAction(PythonInteractiveAction):

    def __str__(self):
        fullname = str(self.py_callable)[1:].split(' at ')[0]
        _, _, shortname = fullname.rpartition('.')
        return "Python: %s" % shortname


class DependencyError(RuntimeError):
    pass


def which(program, raise_err=True):
    '''Checks whether the given program (or program path) is valid and
    executable.

    NOTE: Sometimes copypasta is okay! This function came from stackoverflow:

        http://stackoverflow.com/a/377028/5109965

    Args:
        program (str): Either a program name or full path to a program.

    Returns:
        Return the path to the executable or None if not found
    '''
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    if raise_err:
        raise DependencyError('{0} not found; is it installed?'.format(program))
    else:
        return None


def parallel_fasta(input_filename, output_filename, command, n_jobs, pbs=None):

    exc = which('parallel')
    cmd = ['cat', input_filename, '|', exc, '--round-robin', '--pipe', '-L', 2,
           '-N', 10000, '--gnu']
    if pbs is not None:
        cmd.extend(['--sshloginfile', pbs, '--workdir $PWD'])
    else:
        cmd.extend(['-j', n_jobs])
    cmd.extend(['-a', input_filename])

    if isinstance(command, list):
        command = ' '.join(command)
    cmd.extend([command, '>', output_filename])
    return ' '.join(map(str, cmd))


def hidden_fn(fn):
    return '.{0}'.format(fn)



#!/usr/bin/env python
from __future__ import print_function

import pandas as pd

class BestHits(object):

    def __init__(self, comparison_cols=['E'], query_name_col='q_name', 
                 subject_name_col='s_name', query_length_col='q_len',
                 subject_length_col='s_len'):
        '''Build a BestHits object to manage finding best or reciprocal best
        hits.

        Args:
            comparison_cols (list): Columns to compare when determining which
                hit is "best."
            query_name_col (str): The column with the query sequence names.
            subject_name_col (str): The column with the subject suquence names.
            query_length_col (str): The column with the query length.
            subject_length_col (str): The column with the subject length.
        '''

        self.comparison_cols = comparison_cols
        self.query_name_col = query_name_col
        self.subject_name_col = subject_name_col
        self.query_length_col = query_length_col
        self.subject_length_col = subject_length_col

    def best_hits(self, aln_df, inplace=True):
        '''Get the best hit for each query in the alignment DataFrame.

        Operates in-place. Sorts the hits by query name and then comparison_cols,
        then uses the drop_duplicates() function to remove all but the
        first hit for each query.

        Args:
            aln_df (DataFrame): The MAF alignment DataFrame.
            inplace (bool): If True, perform the operation in-place and
                return the same DataFrame. If False, return a copy.
        Returns:
            DataFrame with the best hits.
        '''

        if inplace:
            aln_df.sort_values(by=[self.query_name_col] + self.comparison_cols,
                               inplace=True)
            aln_df.drop_duplicates(subset=self.query_name_col, inplace=True)
            return aln_df
        else:
            return aln_df.sort_values(
                       by=[self.query_name_col] + self.comparison_cols
                   ).drop_duplicates(subset=self.query_name_col)

    def reciprocal_best_hits(self, aln_df_A, aln_df_B, inplace=False, drop=True):
        '''Given to DataFrames with reciprocal MAF alignments, get the
        reciprocal best hits.

        Uses the best_hits function to get best hits for each DataFrame,
        and then does an inner join to find the reciprocals.

        Args:
            aln_df_A (DataFrame): The query hits.
            aln_df_B (DataFrame): The subject hits.
            inplace (bool): Passed to the best_hits calls.
            drop (bool): Drop extraneous columns and rename.
        Returns:
            DataFrame with the reciprocal best hits.
        '''

        aln_df_A = self.best_hits(aln_df_A, inplace=inplace)
        aln_df_B = self.best_hits(aln_df_B, inplace=inplace)

        # Join between subject A and query B
        rbh_df = pd.merge(aln_df_A, aln_df_B, how='inner', 
                          left_on=self.subject_name_col, 
                          right_on=self.query_name_col,
                          suffixes=('_A', '_B'))

        # Select those where query A is the same as subject B
        rbh_df = rbh_df[(rbh_df[self.query_name_col+'_A'] == rbh_df[self.subject_name_col+'_B'])]

        # Renamed columns after join
        if drop:
            rename_d = {}
            for rbh_col in rbh_df.columns:
                trimmed_col, _, _ = rbh_col.rpartition('_')
                if rbh_col.endswith('_A'):
                    rename_d[rbh_col] = trimmed_col

            rbh_df.rename(columns=rename_d, inplace=True)

            return rbh_df[list(rename_d.values()) + ['q_frame']]
        else:
            return rbh_df





import csv
import filelock
from functools import wraps
import sys
import time
import warnings
from contextlib import contextmanager
from os import path

from doit.action import PythonAction
from doit.task import Task as DoitTask
import six


def cleaned_actions(actions):
    '''Get a cleanup list of actions: Python actions have their <locals> portion
    stripped, which clutters up PythonActions that are closures.
    '''
    txt = ''
    for action in actions:
        txt_rep = six.text_type(action)
        if isinstance(action, PythonAction):
            # clean up inner fuctions in Python actions
            txt_rep = txt_rep.replace('<locals>.', '')
        else:
            txt_rep = txt_rep[:5] + '`' + txt_rep[5:] + '`'
        txt += "\n    * {0}".format(txt_rep)
    return txt


class Profiler(object):
    '''Thread-safe performance profiler.
    '''

    def __init__(self):
        self.running = False

    def start_profiler(self, filename=None, blockname='__main__'):
        '''Start the profiler, with results stored in the given filename.

        Args:
            filename (str): Path to store profiling results. If not given,
                uses a representation of the current time
            blockname (str): Name assigned to the main block.
        '''

        self.run_name = time.strftime("%a_%d_%b_%Y_%H%M%S", time.localtime())
        if filename is None:
            self.filename = '{0}.csv'.format(self.run_name)
        else:
            self.filename = filename
        self.run_name = time.ctime()
        self.start_time = time.time()
        self.blockname = blockname
        self.running = True
        self.lock = filelock.FileLock('{0}.lock'.format(self.filename))
        print('Profiling is ON:', self.filename, '\n', file=sys.stderr)

    def write_result(self, task_name, start_time, end_time, elapsed_time):
        '''Write results to the file, using the given task name as the
        name for the results block.

        Args:
            task_name (str): ID for the result row (the block profiled).
            start_time (float): Time of block start.
            end_time (float): Time of block end.
            elapsed_time (float): Total time.
        '''

        try:
            with self.lock.acquire(timeout=10):
                header = not path.isfile(self.filename)
                with open(self.filename, 'a') as fp:
                    writer = csv.writer(fp, delimiter=',')
                    if header:
                        writer.writerow(['run_id', 'block', 'start_t', 'end_t',
                                      'elapsed_t'])
                    row = [self.run_name, task_name, start_time, end_time, elapsed_time]
                    writer.writerow(row)
        except filelock.Timeout as e:
            warnings.warn(e, RuntimeWarning, stacklevel=1)

    def stop_profiler(self):
        '''Shut down the profiler and write the final elapsed time.
        '''
        self.end_time = time.time()
        elapsed = self.end_time - self.start_time
        self.write_result(self.blockname, self.start_time, self.end_time, elapsed)
        self.running = False
        return elapsed


class Timer(object):
    '''Simple timer class.
    '''

    def start(self):
        '''Start the timer.
        '''
        self.start_time = time.time()

    def stop(self):
        '''Stop the timer and return the elapsed time.
        '''
        self.end_time = time.time()
        return self.end_time - self.start_time


def title_without_profile_actions(task):
    """Generate title without profiling actions"""
    title = ''
    if task.actions:
        title = cleaned_actions(task.actions[1:-1])
    else:
        title = "Group: %s" % ", ".join(task.task_dep)
    return "%s: %s"% (task.name, title)


def setup_profiler():

    profiler = Profiler()
    @contextmanager
    def profiler_manager(filename=None, blockname='__main__'):
        profiler.start_profiler(filename=filename, blockname=blockname)
        yield
        profiler.stop_profiler()

    def add_profile_actions(task):
        timer = Timer()
        def start_profiling():
            if profiler.running:
                timer.start()

        def stop_profiling():
            if profiler.running:
                elapsed = timer.stop()
                profiler.write_result(task['name'], timer.start_time,
                                      timer.end_time, elapsed)
        
        if isinstance(task, DoitTask):
            actions = task._actions
            task.custom_title = title_without_profile_actions
        else:
            actions = task['actions']
            task['title'] = title_without_profile_actions

        actions.insert(0, start_profiling)
        actions.append(stop_profiling)

        return task
    
    def profile_decorator(task_func):

        @wraps(task_func)
        def func(*args, **kwargs):
            task = task_func(*args, **kwargs)
            return add_profile_actions(task)
        
        return func

    return profiler_manager, profile_decorator

StartProfiler, profile_task = setup_profiler()




from os import path, getcwd, mkdir

from doit.tools import run_once, create_folder
from doit.task import clean_targets, dict_to_task
from doit.cmd_base import TaskLoader
from doit.doit_cmd import DoitMain
import pandas as pd

from .crbl import (get_reciprocal_best_last_translated, backmap_names,
                   scale_evalues, fit_crbh_model, filter_hits_from_model,
                   plot_crbh_fit)

from .profile import StartProfiler, profile_task
from .last import lastdb_task, lastal_task, MafParser
from .translate import translate_task, rename_task
from .util import ShortenedPythonAction, title, hidden_fn
from .util import create_doit_task as doit_task


class ShmlastApp(TaskLoader):

    def __init__(self, directory=None, config=None):
        super(ShmlastApp, self).__init__()

        if directory is None:
            directory = getcwd()
        self.directory = directory
        try:
            mkdir(self.directory)
        except OSError:
            pass
        self.doit_config = {'verbosity': 2}
        if config is not None:
            self.doit_config.update(config)

    def tasks(self):
        raise NotImplementedError()

    def load_tasks(self, cmd, opt_values, pos_args):
        return list(self.tasks()), self.doit_config

    def run(self, doit_args=None, move=False, profile_fn=None):
        if doit_args is None:
            doit_args = ['run']
        runner = DoitMain(self)

        print('\n--- Begin Task Execution ---')
        if profile_fn is not False and doit_args[0] == 'run':
            with StartProfiler(filename=profile_fn):
                return runner.run(doit_args)
        else:
            return runner.run(doit_args)


class RBL(ShmlastApp):

    def __init__(self, query_fn, database_fn, output_fn=None,
                 cutoff=.00001, n_threads=1, pbs=None, directory=None):
        '''Generate and manage the pydoit tasks for the RBL pipeline.

        Args:
            query_fn (str): The query filename.
            database_fn (str): The database filename.
            output_fn (str): Filename to store results in.
            cutoff (float): The score cutoff.
            n_threads (int): Number of threads to run on.
            pbs (bool): Whether to generate parallel command for PBS
                environment.
            directory (str): The directory to run tasks in.
        '''

        self.query_fn = query_fn
        self.renamed_query_fn = hidden_fn(path.basename(self.query_fn))
        self.query_name_map_fn = self.renamed_query_fn + '.names.csv'
        self.translated_query_fn = self.renamed_query_fn + '.pep'

        self.database_fn = database_fn
        self.renamed_database_fn = hidden_fn(path.basename(self.database_fn))
        self.database_name_map_fn = self.renamed_database_fn + '.names.csv'

        self.n_threads = n_threads
        self.pbs = pbs
        self.cutoff = cutoff

        self.db_x_query_fn = '{0}.x.{1}.maf'.format(self.renamed_database_fn,
                                                    self.translated_query_fn.strip('.'))

        self.query_x_db_fn = '{0}.x.{1}.maf'.format(self.translated_query_fn,
                                                    self.renamed_database_fn.strip('.'))
        
        self.output_fn = output_fn
        if self.output_fn is None:
            self.output_fn = '{q}.x.{d}.rbl.csv'.format(q=path.basename(self.query_fn),
                                                        d=path.basename(self.database_fn))
        self.unmapped_output_fn = hidden_fn(self.output_fn)
        
        dep_file = '.{0}.shmlast.doit'.format(path.basename(self.query_fn))
        super(RBL, self).__init__(directory=directory, 
                                  config={'dep_file': dep_file})

    @doit_task
    @profile_task
    def reciprocal_best_last_task(self):
       
        def do_reciprocals():
            rbh_df, qvd_df, dvq_df = get_reciprocal_best_last_translated(self.query_x_db_fn,
                                                                         self.db_x_query_fn)
            q_names = pd.read_csv(self.query_name_map_fn)
            d_names = pd.read_csv(self.database_name_map_fn)

            rbh_df.to_csv(self.unmapped_output_fn, index=False)
            rbh_df = backmap_names(rbh_df, q_names, d_names)
            rbh_df.to_csv(self.output_fn, index=False)

        td = {'name': 'reciprocal_best_last',
              'title': title,
              'actions': [ShortenedPythonAction(do_reciprocals)],
              'file_dep': [self.query_x_db_fn,
                           self.db_x_query_fn,
                           self.query_name_map_fn,
                           self.database_name_map_fn],
              'targets': [self.unmapped_output_fn,
                          self.output_fn],
              'clean': [clean_targets]}
        
        return td

    def rename_transcriptome_task(self):
        return rename_task(self.query_fn,
                           self.renamed_query_fn,
                           name_map_fn=self.query_name_map_fn)

    def rename_database_task(self):
        return rename_task(self.database_fn,
                           self.renamed_database_fn,
                           prefix='db',
                           name_map_fn=self.database_name_map_fn)

    def translate_task(self):
        return translate_task(self.renamed_query_fn,
                              self.translated_query_fn)

    def format_transcriptome_task(self):
        return lastdb_task(self.translated_query_fn,
                           prot=True)

    def format_database_task(self):
        return lastdb_task(self.renamed_database_fn,
                           prot=True)

    def align_transcriptome_task(self):
        return lastal_task(self.translated_query_fn,
                           self.renamed_database_fn,
                           self.query_x_db_fn,
                           translate=False, 
                           cutoff=self.cutoff,
                           n_threads=self.n_threads,
                           pbs=self.pbs)

    def align_database_task(self):
        return lastal_task(self.renamed_database_fn,
                           self.translated_query_fn,
                           self.db_x_query_fn,
                           translate=False, 
                           cutoff=self.cutoff,
                           n_threads=self.n_threads,
                           pbs=self.pbs)


    def tasks(self):
        '''Iterator over all tasks in pipeline.
        '''
        yield self.rename_transcriptome_task()
        yield self.rename_database_task()
        yield self.translate_task()
        yield self.format_transcriptome_task()
        yield self.format_database_task()
        yield self.align_database_task()
        yield self.align_transcriptome_task()
        yield self.reciprocal_best_last_task()


class CRBL(RBL):

    def __init__(self, query_fn, database_fn, output_fn=None,
                 model_fn=None, cutoff=.00001, n_threads=1, pbs=None):
        '''Generate and manage the pydoit tasks for the CRBL pipeline.

        Args:
            query_fn (str): The query filename.
            database_fn (str): The database filename.
            output_fn (str): Filename to store results in.
            model_fn (str): Filename to store the model in.
            cutoff (float): The score cutoff.
            n_threads (int): Number of threads to run on.
            pbs (bool): Whether to generate parallel command for PBS
                environment.
            directory (str): The directory to run tasks in.
        '''
        prefix = '{q}.x.{d}.crbl'.format(q=path.basename(query_fn),
                                         d=path.basename(database_fn))

        self.crbl_output_fn = output_fn
        if output_fn is None:
            self.crbl_output_fn = prefix + '.csv'
        self.unmapped_crbl_output_fn = hidden_fn(self.crbl_output_fn)

        self.model_fn = model_fn
        if model_fn is None:
            self.model_fn = prefix + '.model.csv'
            self.model_plot_fn = prefix + '.model.plot.pdf'
        else:
            self.model_plot_fn = self.model_fn + '.plot.pdf'

        super(CRBL, self).__init__(query_fn,
                                    database_fn,
                                    output_fn=None,
                                    cutoff=cutoff,
                                    n_threads=n_threads,
                                    pbs=pbs)

    @doit_task
    @profile_task
    def crbl_fit_and_filter_task(self):

        def do_crbl_fit_and_filter():
            rbh_df, hits_df, _ = get_reciprocal_best_last_translated(self.query_x_db_fn,
                                                                     self.db_x_query_fn)
            
            model_df = fit_crbh_model(rbh_df)
            model_df.to_csv(self.model_fn, index=False)
            model_df = pd.read_csv(self.model_fn)

            filtered_df = filter_hits_from_model(model_df, rbh_df, hits_df)
            results = pd.concat([rbh_df, filtered_df], axis=0)
            del results['translated_q_name']

            q_names = pd.read_csv(self.query_name_map_fn)
            d_names = pd.read_csv(self.database_name_map_fn)
            
            results = backmap_names(results, q_names, d_names)
            results.to_csv(self.crbl_output_fn, index=False)

            plot_crbh_fit(model_df, hits_df, self.model_plot_fn)

        td = {'name': 'fit_and_filter_crbl_hits',
              'title': title,
              'actions': [ShortenedPythonAction(do_crbl_fit_and_filter)],
              'file_dep': [self.query_x_db_fn,
                           self.query_name_map_fn,
                           self.database_name_map_fn],
              'targets': [self.crbl_output_fn, 
                          self.model_plot_fn,
                          self.model_fn],
              'clean': [clean_targets]}
        
        return td

    def tasks(self):
        '''Iterator over all pipeline tasks.
        '''
        for tsk in super(CRBL, self).tasks():
            if tsk.name != 'reciprocal_best_last':
                yield tsk
        yield self.crbl_fit_and_filter_task()
        



#/usr/bin/env python3

import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
from ficus import FigureManager
import numpy as np
from os import path
import pandas as pd
import seaborn as sns

from .hits import BestHits
from .last import MafParser

def get_reciprocal_best_last_translated(query_maf, database_maf):
    '''Perform Reciprocal Best Hits between the given MAF files.

    Args:
        query_maf (str): The query MAF file.
        database_maf (str): The translated datbase MAF file.
    Returns:
        tuple: DataFrames with the RBH's, query vs database, and database vs
            query hits.
    '''
    bh = BestHits(comparison_cols=['E', 'EG2'])
    qvd_df = MafParser(query_maf).read()
    qvd_df[['qg_name', 'q_frame']] = qvd_df.q_name.str.partition('_')[[0,2]]
    qvd_df.rename(columns={'q_name': 'translated_q_name',
                           'qg_name': 'q_name'},
                  inplace=True)
    qvd_df['ID'] = qvd_df.index

    dvq_df = MafParser(database_maf).read()
    dvq_df[['sg_name', 'frame']] = dvq_df.s_name.str.partition('_')[[0,2]]
    dvq_df.rename(columns={'s_name': 'translated_s_name',
                           'sg_name': 's_name'},
                  inplace=True)
    dvq_df['ID'] = dvq_df.index
    
    return bh.reciprocal_best_hits(qvd_df, dvq_df), qvd_df, dvq_df


def backmap_names(results_df, q_names, d_names):
    '''Map names from translated RBH's to original query and database names.

    Args:
        results_df (pandas.DataFrame): The results to backmap.
        q_names (pandas.DataFrame): Query name map.
        d_names (pandas.DataFrame): Database name map.
    Returns:
        pandas.DataFrame: Reference to results_df.
    '''

    results_df = pd.merge(results_df, 
                          q_names, 
                          left_on='q_name',
                          right_on='new_name')
    results_df['q_name'] = results_df['old_name']
    del results_df['old_name']

    results_df = pd.merge(results_df, 
                          d_names, 
                          left_on='s_name',
                          right_on='new_name')
    results_df['s_name'] = results_df['old_name']
    del results_df['old_name']
    del results_df['new_name_x']
    del results_df['new_name_y']

    return results_df


def scale_evalues(df, name='E', inplace=False):
    '''Log scale the evalue column specified by name.

    Args:
        df (pandas.DataFrame): The data.
        name (str): Column name with the evalues.
        inplace (bool): Perform the scaling inplace.
    Returns:
        tuple: The scaled DataFrame and the new column name of the scaled
            values.
    '''

    scaled_col_name = name + '_scaled'
    if inplace is False:
        df = df.copy()
    df[scaled_col_name] = df[name]
    df.loc[df[scaled_col_name] == 0.0, scaled_col_name] = 1e-300
    df[scaled_col_name] = -np.log10(df[scaled_col_name])
    return df, scaled_col_name


def fit_crbh_model(rbh_df, length_col='s_aln_len', feature_col='E'):
    '''Build the CRBH model on the given RBH's.

    Args:
        rbh_df (pandas.DataFrame): DataFrame with RBH's.
        length_col (str): The column with the subject lengths.
        feature_col (str): Score column to train on.
    Returns:
        pandas.DataFrame: The model.
    '''

    data = rbh_df[[length_col, feature_col]].rename(columns={length_col:'length'})
    data.sort_values('length', inplace=True)
    _, feature_col = scale_evalues(data, name=feature_col, inplace=True)

    # create a DataFrame for the model, staring with the alignment lengths
    fit = pd.DataFrame(np.arange(10, data['length'].max()), 
                       columns=['center'], dtype=int)
    
    # create the bins
    fit['size'] = fit['center'] * 0.1
    fit.loc[fit['size'] < 5, 'size'] = 5
    fit['size'] = fit['size'].astype(int)
    fit['left'] = fit['center'] - fit['size']
    fit['right'] = fit['center'] + fit['size']
    
    # do the fitting: it's just a sliding window with an increasing size
    def bin_mean(fit_row, df):
        hits = df[(df['length'] >= fit_row.left) & (df['length'] <= fit_row.right)]
        return hits[feature_col].mean()
    fit['fit'] = fit.apply(bin_mean, args=(data,), axis=1)
    model_df = fit.dropna()

    return model_df


def filter_hits_from_model(model_df, rbh_df, hits_df, feature_col='E',
                           id_col='ID', length_col='s_aln_len'):
    '''Filter a DataFrame of LAST best hits using the CRBH model.

    Args:
        model_df (pandas.DataFrame): The CRBH model.
        rbh_df (pandas.DataFrame): The RBH's.
        hits_df (pandas.DataFrame): The query vs database hits.
        feature_col (str): Column name of scores.
        id_col (str): Column with unique ID of hits.
        length_col (str): Column name to use for length.
    Returns:
        pandas.DataFrame: The CRBH's.
    '''

    hits_df, _ = scale_evalues(hits_df, name=feature_col, inplace=False)
    rbh_df, scaled_feature_col = scale_evalues(rbh_df, name=feature_col, inplace=False)

    # Merge the model into the subset of the hits which aren't in RBH
    comp_df = pd.merge(hits_df[hits_df[id_col].isin(rbh_df[id_col]) == False], 
                       model_df, left_on=length_col, right_on='center')

    crbl_df = comp_df[comp_df[scaled_feature_col] >= comp_df['fit']]

    del crbl_df['center']
    del crbl_df['left']
    del crbl_df['right']
    del crbl_df['fit']
    del crbl_df['size']

    return crbl_df


def plot_crbh_fit(model_df, hits_df, model_plot_fn, show=False,
                  figsize=(10,10), feature_col='E', length_col='s_aln_len'):

    plt.style.use('seaborn-ticks')

    with FigureManager(model_plot_fn, show=show, 
                       figsize=figsize) as (fig, ax):

        scatter_kws = {'s': 10, 'alpha':0.7}
        scatter_kws['c'] = sns.xkcd_rgb['ruby']
        scatter_kws['marker'] = 'o'
        line_kws = {'c': sns.xkcd_rgb['red wine'], 
                    'label':'Query Hits Regression'}
        sample_size = min(len(hits_df), 5000)
        hits_df, scaled_col = scale_evalues(hits_df, name=feature_col,
                                            inplace=False)
        sns.regplot(length_col, scaled_col, hits_df.sample(sample_size), order=1, 
                    label='Query Hits', scatter_kws=scatter_kws, 
                    line_kws=line_kws, color=scatter_kws['c'], ax=ax)

        scatter_kws['c'] = sns.xkcd_rgb['twilight blue']
        scatter_kws['marker'] = 's'
        sns.regplot('center', 'fit', model_df, 
                    fit_reg=False, x_jitter=True, y_jitter=True, ax=ax,
                    label='CRBL Fit', scatter_kws=scatter_kws, line_kws=line_kws)

        leg = ax.legend(fontsize='medium', scatterpoints=3, frameon=True)
        leg.get_frame().set_linewidth(1.0)

        ax.set_xlim(model_df['center'].min(), model_df['center'].max())
        ax.set_ylim(0, max(model_df['fit'].max(), hits_df[scaled_col].max()) + 50)
        ax.set_ylabel('Score ({0})'.format(feature_col))
        ax.set_xlabel('Alignment Length')





from __future__ import unicode_literals
import os
from pkg_resources import Requirement, resource_filename, ResolutionError
import shutil
import stat
import sys
from tempfile import mkdtemp
import traceback

from distutils import dir_util
from pytest import fixture


from doit.cmd_base import TaskLoader
from doit.doit_cmd import DoitMain
from doit.dependency import Dependency, DbmDB

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

@fixture
def datadir(tmpdir, request):
    '''
    Fixture responsible for locating the test data directory and copying it
    into a temporary directory.
    '''
    filename = request.module.__file__
    test_dir = os.path.dirname(filename)
    data_dir = os.path.join(test_dir, 'data') 
    dir_util.copy_tree(data_dir, str(tmpdir))

    def getter(filename, as_str=True):
        filepath = tmpdir.join(filename)
        if as_str:
            return str(filepath)
        return filepath

    return getter


def check_status(task, tasks=None, dep_file='.doit.db'):
    if tasks is None:
        tasks = [task]
    mgr = Dependency(DbmDB, os.path.abspath(dep_file))
    status = mgr.get_status(task, tasks)
    return status


def run_tasks(tasks, args, config={'verbosity': 0}):
    
    if type(tasks) is not list:
        raise TypeError('tasks must be a list')
   
    class Loader(TaskLoader):
        @staticmethod
        def load_tasks(cmd, opt_values, pos_args):
            return tasks, config
   
    return DoitMain(Loader()).run(args)


def run_task(task, cmd='run', verbosity=0):
    return run_tasks([task], [cmd], config={'verbosity': verbosity})


def touch(filename):
    '''Perform the equivalent of bash's touch on the file.

    Args:
        filename (str): File path to touch.
    '''

    open(filename, 'a').close()
    os.chmod(filename, stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)



'''
These script running functions were taken from the khmer project:
https://github.com/dib-lab/khmer/blob/master/tests/khmer_tst_utils.py
'''

def scriptpath(scriptname='shmlast'):
    "Return the path to the scripts, in both dev and install situations."

    path = os.path.join(os.path.dirname(__file__), "../../bin")
    if os.path.exists(os.path.join(path, scriptname)):
        return path

    path = os.path.join(os.path.dirname(__file__), "../../../EGG-INFO/bin")
    if os.path.exists(os.path.join(path, scriptname)):
        return path

    for path in os.environ['PATH'].split(':'):
        if os.path.exists(os.path.join(path, scriptname)):
            return path


def _runscript(scriptname):
    """
    Find & run a script with exec (i.e. not via os.system or subprocess).
    """

    import pkg_resources
    ns = {"__name__": "__main__"}
    ns['sys'] = globals()['sys']

    try:
        pkg_resources.get_distribution("shmlast").run_script(scriptname, ns)
        return 0
    except pkg_resources.ResolutionError as err:
        path = scriptpath()

        scriptfile = os.path.join(path, scriptname)
        if os.path.isfile(scriptfile):
            if os.path.isfile(scriptfile):
                exec(compile(open(scriptfile).read(), scriptfile, 'exec'), ns)
                return 0

    return -1


def runscript(scriptname, args, directory=None,
              fail_ok=False, sandbox=False):
    """Run a Python script using exec().
    Run the given Python script, with the given args, in the given directory,
    using 'exec'.  Mimic proper shell functionality with argv, and capture
    stdout and stderr.
    When using :attr:`fail_ok`=False in tests, specify the expected error.
    """
    sysargs = [scriptname]
    sysargs.extend(args)
    cwd = os.getcwd()

    try:
        status = -1
        oldargs = sys.argv
        sys.argv = sysargs

        oldout, olderr = sys.stdout, sys.stderr
        sys.stdout = StringIO()
        sys.stdout.name = "StringIO"
        sys.stderr = StringIO()

        if directory:
            os.chdir(directory)
        else:
            directory = cwd

        try:
            print('running:', scriptname, 'in:', directory, file=oldout)
            print('arguments', sysargs, file=oldout)

            status = _runscript(scriptname)
        except SystemExit as e:
            status = e.code
        except:
            traceback.print_exc(file=sys.stderr)
            status = -1
    finally:
        sys.argv = oldargs
        out, err = sys.stdout.getvalue(), sys.stderr.getvalue()
        sys.stdout, sys.stderr = oldout, olderr

        os.chdir(cwd)

    if status != 0 and not fail_ok:
        print('Script Failed:', scriptname, 
              'Status:', status, 
              'Output:', out,
              'Error:', err,
              sep='\n')
        assert False

    return status, out, err


def run_shell_cmd(cmd, fail_ok=False, in_directory=None):
    cwd = os.getcwd()
    if in_directory:
        os.chdir(in_directory)

    print('running: ', cmd)
    try:
        p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        (out, err) = p.communicate()

        out = out.decode('utf-8')
        err = err.decode('utf-8')

        if p.returncode != 0 and not fail_ok:
            print('out:', out)
            print('err:', err)
            raise AssertionError("exit code is non zero: %d" % p.returncode)

        return (p.returncode, out, err)
    finally:
        os.chdir(cwd)




#!/usr/bin/env python3

import sys, platform, os

try:
    from setuptools import *
except ImportError:
    from distribute_setup import use_setuptools
    use_setuptools()
finally:
    from setuptools import *

from glob import glob

if sys.version_info < (3, 4):
    print >> sys.stderr, "ERROR: shmlast requires python 3.4 or greater"
    sys.exit()

__version__ = open(os.path.join('shmlast', 'VERSION')).read().strip()

SCRIPTS=glob('bin/*')

def main():
    setup(  name = 'shmlast',
            version = __version__,
            description = 'An improved implementation of Conditional Reciprocal'\
                          ' Best Hits with LAST and Python.',
            long_description = open("README.rst").read(),
            url = 'https://github.com/camillescott/shmlast',
            author = 'Camille Scott',
            author_email = 'camille.scott.w@gmail.com',
            license = 'BSD',
            packages = find_packages(),
            scripts = SCRIPTS,
            setup_requires = ['pytest-runner'],
            tests_require = ['pytest',
                             'codecov'],
            install_requires = ['doit>=0.29.0',
                                'ficus>=0.3.2',
                                'matplotlib>=1.4',
                                'numpy>=1.9.0',
                                'pandas>=0.17.0',
                                'scipy>=0.16.0',
                                'screed>=0.9',
                                'seaborn>=0.6.0',
                                'pytest>=2.5',
                                'filelock>=2.0.6'],
            zip_safe = False,
            include_package_data = True )
            

if __name__ == "__main__":
    main()



from doit.task import clean_targets
import pandas as pd
import screed

from .profile import profile_task
from .util import create_doit_task as doit_task
from .util import ShortenedPythonAction, title, which


dna_to_aa={'TTT':'F','TTC':'F', 'TTA':'L','TTG':'L',
                'TCT':'S','TCC':'S','TCA':'S','TCG':'S',
                'TAT':'Y','TAC':'Y', 'TAA':'X','TAG':'X','TGA':'X',
                'TGT':'C','TGC':'C', 'TGG':'W',
                'CTT':'L','CTC':'L','CTA':'L','CTG':'L',
                'CCT':'P','CCC':'P','CCA':'P','CCG':'P',
                'CAT':'H','CAC':'H', 'CAA':'Q','CAG':'Q',
                'CGT':'R','CGC':'R','CGA':'R','CGG':'R',
                'ATT':'I','ATC':'I','ATA':'I', 'ATG':'M',
                'ACT':'T','ACC':'T','ACA':'T','ACG':'T',
                'AAT':'N','AAC':'N', 'AAA':'K','AAG':'K',
                'AGT':'S','AGC':'S', 'AGA':'R','AGG':'R',
                'GTT':'V','GTC':'V','GTA':'V','GTG':'V',
                'GCT':'A','GCC':'A','GCA':'A','GCG':'A',
                'GAT':'D','GAC':'D', 'GAA':'E','GAG':'E',
                'GGT':'G','GGC':'G','GGA':'G','GGG':'G'}


__complementTranslation = { "A": "T", "C": "G", "G": "C", "T": "A", "N": "N" }
def complement(s):
    ''' Get the base complement.

    Args:
        s (str): The sequence to complement.
    '''
    c = "".join(__complementTranslation[n] for n in s)
    return c


def reverse(s):
    '''Reverse the sequence.

    Args:
        s (str): Sequence to reverse.
    '''
    r = "".join(reversed(s))
    return r


def peptides(seq, start):
    '''Translate the nucleotide sequence.

    Args:
        seq (str): The nucleotide sequence.
        start (int): Translation start position.
    Yields:
        A sequence of amino acid identifiers for each codon.
    '''

    for i in range(start, len(seq), 3):
        yield dna_to_aa.get(seq[i:i+3], "X")


def translate(seq):
    '''6-frame translation of the given nucleotide sequence.

    Args:
        seq (str): The nucleotide sequence.
    Yields:
        str: The translation in each frame.
    '''

    for i in range(3):
        pep = peptides(seq, i)
        yield "".join(pep)

    revcomp = reverse(complement((seq)))
    for i in range(3):
        pep = peptides(revcomp, i)
        yield "".join(pep)


def translate_fastx(input_fn, output_fn):
    '''Translate a nucleotide FASTA file.

    Args:
        input_fn (str): The FASTA file to translate.
        output_fn (str): Filename to store the results.
    '''

    with open(output_fn, 'w') as fp:
        for record in screed.open(input_fn):
            for frame, t in enumerate(translate(record.sequence)):
                name = '{0}_{1}'.format(record.name, frame)
                fp.write('>{0}\n{1}\n'.format(name, t))


@doit_task
@profile_task
def rename_task(input_fn, output_fn, name_map_fn='name_map.csv', prefix='tr'):
    '''Rename the FASTA idenfiers to play nicely with various programs.

    Args:
        input_fn (str): The FASTA to rename.
        output_fn (str): The filename of the renamed version.
        name_map_fn (str): Where to store the mapping of old to new names.
        prefix (str): Prefix to use for each transcript.
    Returns:
        dict: A doit task dictionary.
    '''
    
    def rename_input():
        name_map = []
        with open(output_fn, 'w') as output_fp:
            for n, record in enumerate(screed.open(input_fn)):
                new_name = '{0}{1}'.format(prefix, n)
                output_fp.write('>{0}\n{1}\n'.format(new_name,
                                                     record.sequence))
                name_map.append((record.name, new_name))

        pd.DataFrame(name_map,
                     columns=['old_name', 'new_name']).to_csv(name_map_fn,
                                                              index=False)

    return {'name': 'rename:{0}'.format(input_fn),
            'title': title,
            'actions': [ShortenedPythonAction(rename_input)],
            'targets': [output_fn, name_map_fn],
            'file_dep': [input_fn],
            'clean': [clean_targets]}


@doit_task
@profile_task
def translate_task(input_fn, output_fn): 
    '''Translate a nucleotide FASTA in six frames.

    Args:
        input_fn (str): The nucleotide FASTA.
        output_fn (str): Destination translated FASTA.
    Returns:
        dict: A doit task dictionary.
    '''

    return {'name': 'translate:{0}'.format(input_fn),
            'title': title,
            'actions': [ShortenedPythonAction(translate_fastx, args=[input_fn, output_fn])],
            'targets': [output_fn],
            'file_dep': [input_fn],
            'clean': [clean_targets]}

